@@ -59,7 +59,7 @@ pub enum ErrorKind {
     FormatSignatureInvalidSignature,
     FormatSealedSignature,
     FormatEmptyKeys,
-    FormatUnknownPublickKey,
+    FormatUnknownPublicKey,
     FormatDeserializationError,
     FormatSerializationError,
     FormatBlockDeserializationError,
@@ -74,15 +74,18 @@ pub enum ErrorKind {
     LogicInvalidAmbientFact,
     LogicInvalidBlockFact,
     LogicInvalidBlockRule,
-    LogicFailedChecks,
-    LogicVerifierNotEmpty,
-    LogicDeny,
+    LogicUnauthorized,
+    LogicAuthorizerNotEmpty,
     LogicNoMatchingPolicy,
-    ParseError,
+    LanguageError,
     TooManyFacts,
     TooManyIterations,
     Timeout,
     ConversionError,
+    FormatInvalidKeySize,
+    FormatInvalidSignatureSize,
+    FormatInvalidKey,
+    FormatSignatureInvalidSignatureGeneration,
 }
 
 #[no_mangle]
@@ -97,12 +100,15 @@ pub extern "C" fn error_kind() -> ErrorKind {
                     Token::Format(Format::Signature(Signature::InvalidFormat)) => {
                         ErrorKind::FormatSignatureInvalidFormat
                     }
-                    Token::Format(Format::Signature(Signature::InvalidSignature)) => {
+                    Token::Format(Format::Signature(Signature::InvalidSignature(_))) => {
                         ErrorKind::FormatSignatureInvalidSignature
                     }
+                    Token::Format(Format::Signature(Signature::InvalidSignatureGeneration(_))) => {
+                        ErrorKind::FormatSignatureInvalidSignatureGeneration
+                    }
                     Token::Format(Format::SealedSignature) => ErrorKind::FormatSealedSignature,
                     Token::Format(Format::EmptyKeys) => ErrorKind::FormatEmptyKeys,
-                    Token::Format(Format::UnknownPublicKey) => ErrorKind::FormatUnknownPublickKey,
+                    Token::Format(Format::UnknownPublicKey) => ErrorKind::FormatUnknownPublicKey,
                     Token::Format(Format::DeserializationError(_)) => {
                         ErrorKind::FormatDeserializationError
                     }
@@ -116,12 +122,17 @@ pub extern "C" fn error_kind() -> ErrorKind {
                         ErrorKind::FormatBlockSerializationError
                     }
                     Token::Format(Format::Version { .. }) => ErrorKind::FormatVersion,
+                    Token::Format(Format::InvalidKeySize(_)) => ErrorKind::FormatInvalidKeySize,
+                    Token::Format(Format::InvalidSignatureSize(_)) => {
+                        ErrorKind::FormatInvalidSignatureSize
+                    }
+                    Token::Format(Format::InvalidKey(_)) => ErrorKind::FormatInvalidKey,
                     Token::InvalidAuthorityIndex(_) => ErrorKind::InvalidAuthorityIndex,
                     Token::InvalidBlockIndex(_) => ErrorKind::InvalidBlockIndex,
                     Token::SymbolTableOverlap => ErrorKind::SymbolTableOverlap,
                     Token::MissingSymbols => ErrorKind::MissingSymbols,
                     Token::Sealed => ErrorKind::Sealed,
-                    Token::ParseError => ErrorKind::ParseError,
+                    Token::Language(_) => ErrorKind::LanguageError,
                     Token::FailedLogic(Logic::InvalidAuthorityFact(_)) => {
                         ErrorKind::LogicInvalidAuthorityFact
                     }
@@ -134,10 +145,13 @@ pub extern "C" fn error_kind() -> ErrorKind {
                     Token::FailedLogic(Logic::InvalidBlockRule(_, _)) => {
                         ErrorKind::LogicInvalidBlockRule
                     }
-                    Token::FailedLogic(Logic::FailedChecks(_)) => ErrorKind::LogicFailedChecks,
-                    Token::FailedLogic(Logic::VerifierNotEmpty) => ErrorKind::LogicVerifierNotEmpty,
-                    Token::FailedLogic(Logic::Deny(_)) => ErrorKind::LogicDeny,
-                    Token::FailedLogic(Logic::NoMatchingPolicy) => ErrorKind::LogicNoMatchingPolicy,
+                    Token::FailedLogic(Logic::Unauthorized { .. }) => ErrorKind::LogicUnauthorized,
+                    Token::FailedLogic(Logic::AuthorizerNotEmpty) => {
+                        ErrorKind::LogicAuthorizerNotEmpty
+                    }
+                    Token::FailedLogic(Logic::NoMatchingPolicy { .. }) => {
+                        ErrorKind::LogicNoMatchingPolicy
+                    }
                     Token::RunLimit(RunLimit::TooManyFacts) => ErrorKind::TooManyFacts,
                     Token::RunLimit(RunLimit::TooManyIterations) => ErrorKind::TooManyIterations,
                     Token::RunLimit(RunLimit::Timeout) => ErrorKind::Timeout,
@@ -154,7 +168,10 @@ pub extern "C" fn error_kind() -> ErrorKind {
 pub extern "C" fn error_check_count() -> u64 {
     use crate::error::*;
     LAST_ERROR.with(|prev| match *prev.borrow() {
-        Some(Error::Biscuit(Token::FailedLogic(Logic::FailedChecks(ref v)))) => v.len() as u64,
+        Some(Error::Biscuit(Token::FailedLogic(Logic::Unauthorized { ref checks, .. })))
+        | Some(Error::Biscuit(Token::FailedLogic(Logic::NoMatchingPolicy { ref checks }))) => {
+            checks.len() as u64
+        }
         _ => 0,
     })
 }
@@ -163,13 +180,16 @@ pub extern "C" fn error_check_count() -> u64 {
 pub extern "C" fn error_check_id(check_index: u64) -> u64 {
     use crate::error::*;
     LAST_ERROR.with(|prev| match *prev.borrow() {
-        Some(Error::Biscuit(Token::FailedLogic(Logic::FailedChecks(ref v)))) => {
-            if check_index >= v.len() as u64 {
+        Some(Error::Biscuit(Token::FailedLogic(Logic::Unauthorized { ref checks, .. })))
+        | Some(Error::Biscuit(Token::FailedLogic(Logic::NoMatchingPolicy { ref checks }))) => {
+            if check_index >= checks.len() as u64 {
                 u64::MAX
             } else {
-                match v[check_index as usize] {
+                match checks[check_index as usize] {
                     FailedCheck::Block(FailedBlockCheck { check_id, .. }) => check_id as u64,
-                    FailedCheck::Verifier(FailedVerifierCheck { check_id, .. }) => check_id as u64,
+                    FailedCheck::Authorizer(FailedAuthorizerCheck { check_id, .. }) => {
+                        check_id as u64
+                    }
                 }
             }
         }
@@ -181,11 +201,12 @@ pub extern "C" fn error_check_id(check_index: u64) -> u64 {
 pub extern "C" fn error_check_block_id(check_index: u64) -> u64 {
     use crate::error::*;
     LAST_ERROR.with(|prev| match *prev.borrow() {
-        Some(Error::Biscuit(Token::FailedLogic(Logic::FailedChecks(ref v)))) => {
-            if check_index >= v.len() as u64 {
+        Some(Error::Biscuit(Token::FailedLogic(Logic::Unauthorized { ref checks, .. })))
+        | Some(Error::Biscuit(Token::FailedLogic(Logic::NoMatchingPolicy { ref checks }))) => {
+            if check_index >= checks.len() as u64 {
                 u64::MAX
             } else {
-                match v[check_index as usize] {
+                match checks[check_index as usize] {
                     FailedCheck::Block(FailedBlockCheck { block_id, .. }) => block_id as u64,
                     _ => u64::MAX,
                 }
@@ -205,13 +226,14 @@ pub extern "C" fn error_check_rule(check_index: u64) -> *const c_char {
     }
 
     LAST_ERROR.with(|prev| match *prev.borrow() {
-        Some(Error::Biscuit(Token::FailedLogic(Logic::FailedChecks(ref v)))) => {
-            if check_index >= v.len() as u64 {
+        Some(Error::Biscuit(Token::FailedLogic(Logic::Unauthorized { ref checks, .. })))
+        | Some(Error::Biscuit(Token::FailedLogic(Logic::NoMatchingPolicy { ref checks }))) => {
+            if check_index >= checks.len() as u64 {
                 std::ptr::null()
             } else {
-                let rule = match &v[check_index as usize] {
+                let rule = match &checks[check_index as usize] {
                     FailedCheck::Block(FailedBlockCheck { rule, .. }) => rule,
-                    FailedCheck::Verifier(FailedVerifierCheck { rule, .. }) => rule,
+                    FailedCheck::Authorizer(FailedAuthorizerCheck { rule, .. }) => rule,
                 };
                 let err = CString::new(rule.clone()).ok();
                 CAVEAT_RULE.with(|ret| {
@@ -228,16 +250,17 @@ pub extern "C" fn error_check_rule(check_index: u64) -> *const c_char {
 }
 
 #[no_mangle]
-pub extern "C" fn error_check_is_verifier(check_index: u64) -> bool {
+pub extern "C" fn error_check_is_authorizer(check_index: u64) -> bool {
     use crate::error::*;
     LAST_ERROR.with(|prev| match *prev.borrow() {
-        Some(Error::Biscuit(Token::FailedLogic(Logic::FailedChecks(ref v)))) => {
-            if check_index >= v.len() as u64 {
+        Some(Error::Biscuit(Token::FailedLogic(Logic::Unauthorized { ref checks, .. })))
+        | Some(Error::Biscuit(Token::FailedLogic(Logic::NoMatchingPolicy { ref checks }))) => {
+            if check_index >= checks.len() as u64 {
                 false
             } else {
-                match v[check_index as usize] {
+                match checks[check_index as usize] {
                     FailedCheck::Block(FailedBlockCheck { .. }) => false,
-                    FailedCheck::Verifier(FailedVerifierCheck { .. }) => true,
+                    FailedCheck::Authorizer(FailedAuthorizerCheck { .. }) => true,
                 }
             }
         }
@@ -250,7 +273,7 @@ pub struct KeyPair(crate::crypto::KeyPair);
 pub struct PublicKey(crate::crypto::PublicKey);
 pub struct BiscuitBuilder<'a>(crate::token::builder::BiscuitBuilder<'a>);
 pub struct BlockBuilder(crate::token::builder::BlockBuilder);
-pub struct Verifier(crate::token::verifier::Verifier);
+pub struct Authorizer<'t>(crate::token::authorizer::Authorizer<'t>);
 
 #[no_mangle]
 pub unsafe extern "C" fn key_pair_new<'a>(
@@ -303,7 +326,7 @@ pub unsafe extern "C" fn key_pair_serialize(kp: Option<&KeyPair>, buffer_ptr: *m
 pub unsafe extern "C" fn key_pair_deserialize(buffer_ptr: *mut u8) -> Option<Box<KeyPair>> {
     let input_slice = std::slice::from_raw_parts_mut(buffer_ptr, 32);
 
-    match crate::crypto::PrivateKey::from_bytes(input_slice) {
+    match crate::crypto::PrivateKey::from_bytes(input_slice).ok() {
         None => {
             update_last_error(Error::InvalidArgument);
             None
@@ -338,7 +361,7 @@ pub unsafe extern "C" fn public_key_serialize(
 pub unsafe extern "C" fn public_key_deserialize(buffer_ptr: *mut u8) -> Option<Box<PublicKey>> {
     let input_slice = std::slice::from_raw_parts_mut(buffer_ptr, 32);
 
-    match crate::crypto::PublicKey::from_bytes(input_slice) {
+    match crate::crypto::PublicKey::from_bytes(input_slice).ok() {
         None => {
             update_last_error(Error::InvalidArgument);
             None
@@ -503,29 +526,18 @@ pub unsafe extern "C" fn biscuit_builder_build<'a>(
 pub unsafe extern "C" fn biscuit_builder_free<'a>(_builder: Option<Box<BiscuitBuilder<'a>>>) {}
 
 #[no_mangle]
-pub unsafe extern "C" fn biscuit_from(
+pub unsafe extern "C" fn biscuit_from<'a>(
     biscuit_ptr: *const u8,
     biscuit_len: usize,
+    root: Option<&'a PublicKey>,
 ) -> Option<Box<Biscuit>> {
     let biscuit = std::slice::from_raw_parts(biscuit_ptr, biscuit_len);
+    if root.is_none() {
+        update_last_error(Error::InvalidArgument);
+    }
+    let root = root?;
 
-    crate::token::Biscuit::from(biscuit)
-        .map(Biscuit)
-        .map(Box::new)
-        .ok()
-}
-
-#[no_mangle]
-pub unsafe extern "C" fn biscuit_from_sealed(
-    biscuit_ptr: *const u8,
-    biscuit_len: usize,
-    secret_ptr: *const u8,
-    secret_len: usize,
-) -> Option<Box<Biscuit>> {
-    let biscuit = std::slice::from_raw_parts(biscuit_ptr, biscuit_len);
-    let secret = std::slice::from_raw_parts(secret_ptr, secret_len);
-
-    crate::token::Biscuit::from_sealed(biscuit, secret)
+    crate::token::Biscuit::from(biscuit, |_| root.0)
         .map(Biscuit)
         .map(Box::new)
         .ok()
@@ -558,7 +570,7 @@ pub unsafe extern "C" fn biscuit_sealed_size(biscuit: Option<&Biscuit>) -> usize
 
     let biscuit = biscuit.unwrap();
 
-    match biscuit.0.sealed_size() {
+    match biscuit.0.serialized_size() {
         Ok(sz) => sz,
         Err(e) => {
             update_last_error(Error::Biscuit(e));
@@ -604,8 +616,6 @@ pub unsafe extern "C" fn biscuit_serialize(
 #[no_mangle]
 pub unsafe extern "C" fn biscuit_serialize_sealed(
     biscuit: Option<&Biscuit>,
-    secret_ptr: *const u8,
-    secret_len: usize,
     buffer_ptr: *mut u8,
 ) -> usize {
     if biscuit.is_none() {
@@ -614,23 +624,29 @@ pub unsafe extern "C" fn biscuit_serialize_sealed(
     }
 
     let biscuit = biscuit.unwrap();
-    let secret = std::slice::from_raw_parts(secret_ptr, secret_len);
 
-    match (*biscuit).0.seal(secret) {
-        Ok(v) => {
-            let size = match biscuit.0.serialized_size() {
-                Ok(sz) => sz,
-                Err(e) => {
-                    update_last_error(Error::Biscuit(e));
-                    return 0;
-                }
-            };
+    match (*biscuit).0.seal() {
+        Ok(b) => match b.to_vec() {
+            Ok(v) => {
+                let size = match biscuit.0.serialized_size() {
+                    Ok(sz) => sz,
+                    Err(e) => {
+                        update_last_error(Error::Biscuit(e));
+                        return 0;
+                    }
+                };
 
-            let output_slice = std::slice::from_raw_parts_mut(buffer_ptr, size);
+                let output_slice = std::slice::from_raw_parts_mut(buffer_ptr, size);
+
+                output_slice.copy_from_slice(&v[..]);
+                v.len()
+            }
+            Err(e) => {
+                update_last_error(Error::Biscuit(e));
+                0
+            }
+        },
 
-            output_slice.copy_from_slice(&v[..]);
-            v.len()
-        }
         Err(e) => {
             update_last_error(Error::Biscuit(e));
             0
@@ -900,8 +916,6 @@ pub unsafe extern "C" fn biscuit_append_block(
     biscuit: Option<&Biscuit>,
     block_builder: Option<&BlockBuilder>,
     key_pair: Option<&KeyPair>,
-    seed_ptr: *const u8,
-    seed_len: usize,
 ) -> Option<Box<Biscuit>> {
     if biscuit.is_none() {
         update_last_error(Error::InvalidArgument);
@@ -918,20 +932,9 @@ pub unsafe extern "C" fn biscuit_append_block(
     }
     let key_pair = key_pair?;
 
-    let slice = std::slice::from_raw_parts(seed_ptr, seed_len);
-    if slice.len() != 32 {
-        update_last_error(Error::InvalidArgument);
-        return None;
-    }
-
-    let mut seed = [0u8; 32];
-    seed.copy_from_slice(slice);
-
-    let mut rng: StdRng = SeedableRng::from_seed(seed);
-
     match biscuit
         .0
-        .append_with_rng(&mut rng, &key_pair.0, builder.0.clone())
+        .append_with_keypair(&key_pair.0, builder.0.clone())
     {
         Ok(token) => Some(Box::new(Biscuit(token))),
         Err(e) => {
@@ -942,25 +945,15 @@ pub unsafe extern "C" fn biscuit_append_block(
 }
 
 #[no_mangle]
-pub unsafe extern "C" fn biscuit_verify<'a, 'b>(
+pub unsafe extern "C" fn biscuit_authorizer<'a, 'b>(
     biscuit: Option<&'a Biscuit>,
-    root: Option<&'b PublicKey>,
-) -> Option<Box<Verifier>> {
+) -> Option<Box<Authorizer<'a>>> {
     if biscuit.is_none() {
         update_last_error(Error::InvalidArgument);
     }
     let biscuit = biscuit?;
-    if root.is_none() {
-        update_last_error(Error::InvalidArgument);
-    }
-    let root = root?;
 
-    (*biscuit)
-        .0
-        .verify((*root).0)
-        .map(Verifier)
-        .map(Box::new)
-        .ok()
+    (*biscuit).0.authorizer().map(Authorizer).map(Box::new).ok()
 }
 
 #[no_mangle]
@@ -1076,15 +1069,15 @@ pub unsafe extern "C" fn block_builder_add_check(
 pub unsafe extern "C" fn block_builder_free(_builder: Option<Box<BlockBuilder>>) {}
 
 #[no_mangle]
-pub unsafe extern "C" fn verifier_add_fact(
-    verifier: Option<&mut Verifier>,
+pub unsafe extern "C" fn authorizer_add_fact(
+    authorizer: Option<&mut Authorizer>,
     fact: *const c_char,
 ) -> bool {
-    if verifier.is_none() {
+    if authorizer.is_none() {
         update_last_error(Error::InvalidArgument);
         return false;
     }
-    let verifier = verifier.unwrap();
+    let authorizer = authorizer.unwrap();
 
     let fact = CStr::from_ptr(fact);
     let s = fact.to_str();
@@ -1093,7 +1086,7 @@ pub unsafe extern "C" fn verifier_add_fact(
         return false;
     }
 
-    verifier
+    authorizer
         .0
         .add_fact(s.unwrap())
         .map_err(|e| {
@@ -1103,15 +1096,15 @@ pub unsafe extern "C" fn verifier_add_fact(
 }
 
 #[no_mangle]
-pub unsafe extern "C" fn verifier_add_rule(
-    verifier: Option<&mut Verifier>,
+pub unsafe extern "C" fn authorizer_add_rule(
+    authorizer: Option<&mut Authorizer>,
     rule: *const c_char,
 ) -> bool {
-    if verifier.is_none() {
+    if authorizer.is_none() {
         update_last_error(Error::InvalidArgument);
         return false;
     }
-    let verifier = verifier.unwrap();
+    let authorizer = authorizer.unwrap();
 
     let rule = CStr::from_ptr(rule);
     let s = rule.to_str();
@@ -1120,7 +1113,7 @@ pub unsafe extern "C" fn verifier_add_rule(
         return false;
     }
 
-    verifier
+    authorizer
         .0
         .add_rule(s.unwrap())
         .map_err(|e| {
@@ -1130,15 +1123,15 @@ pub unsafe extern "C" fn verifier_add_rule(
 }
 
 #[no_mangle]
-pub unsafe extern "C" fn verifier_add_check(
-    verifier: Option<&mut Verifier>,
+pub unsafe extern "C" fn authorizer_add_check(
+    authorizer: Option<&mut Authorizer>,
     check: *const c_char,
 ) -> bool {
-    if verifier.is_none() {
+    if authorizer.is_none() {
         update_last_error(Error::InvalidArgument);
         return false;
     }
-    let verifier = verifier.unwrap();
+    let authorizer = authorizer.unwrap();
 
     let check = CStr::from_ptr(check);
     let s = check.to_str();
@@ -1147,7 +1140,7 @@ pub unsafe extern "C" fn verifier_add_check(
         return false;
     }
 
-    verifier
+    authorizer
         .0
         .add_check(s.unwrap())
         .map_err(|e| {
@@ -1157,14 +1150,41 @@ pub unsafe extern "C" fn verifier_add_check(
 }
 
 #[no_mangle]
-pub unsafe extern "C" fn verifier_verify(verifier: Option<&mut Verifier>) -> bool {
-    if verifier.is_none() {
+pub unsafe extern "C" fn authorizer_add_policy(
+    authorizer: Option<&mut Authorizer>,
+    policy: *const c_char,
+) -> bool {
+    if authorizer.is_none() {
         update_last_error(Error::InvalidArgument);
         return false;
     }
-    let verifier = verifier.unwrap();
+    let authorizer = authorizer.unwrap();
 
-    match verifier.0.verify() {
+    let policy = CStr::from_ptr(policy);
+    let s = policy.to_str();
+    if s.is_err() {
+        update_last_error(Error::InvalidArgument);
+        return false;
+    }
+
+    authorizer
+        .0
+        .add_policy(s.unwrap())
+        .map_err(|e| {
+            update_last_error(Error::Biscuit(e));
+        })
+        .is_ok()
+}
+
+#[no_mangle]
+pub unsafe extern "C" fn authorizer_authorize(authorizer: Option<&mut Authorizer>) -> bool {
+    if authorizer.is_none() {
+        update_last_error(Error::InvalidArgument);
+        return false;
+    }
+    let authorizer = authorizer.unwrap();
+
+    match authorizer.0.authorize() {
         Ok(_index) => true,
         Err(e) => {
             update_last_error(Error::Biscuit(e));
@@ -1174,14 +1194,14 @@ pub unsafe extern "C" fn verifier_verify(verifier: Option<&mut Verifier>) -> boo
 }
 
 #[no_mangle]
-pub unsafe extern "C" fn verifier_print(verifier: Option<&mut Verifier>) -> *mut c_char {
-    if verifier.is_none() {
+pub unsafe extern "C" fn authorizer_print(authorizer: Option<&mut Authorizer>) -> *mut c_char {
+    if authorizer.is_none() {
         update_last_error(Error::InvalidArgument);
         return std::ptr::null_mut();
     }
-    let verifier = verifier.unwrap();
+    let authorizer = authorizer.unwrap();
 
-    match CString::new(verifier.0.print_world()) {
+    match CString::new(authorizer.0.print_world()) {
         Ok(s) => s.into_raw(),
         Err(_) => {
             update_last_error(Error::InvalidArgument);
@@ -1191,7 +1211,7 @@ pub unsafe extern "C" fn verifier_print(verifier: Option<&mut Verifier>) -> *mut
 }
 
 #[no_mangle]
-pub unsafe extern "C" fn verifier_free(_verifier: Option<Box<Verifier>>) {}
+pub unsafe extern "C" fn authorizer_free(_authorizer: Option<Box<Authorizer>>) {}
 
 #[no_mangle]
 pub unsafe extern "C" fn string_free(ptr: *mut c_char) {
