@@ -5,7 +5,7 @@
 //!
 //! ```rust
 //! use std::convert::TryInto;
-//! use biscuit_auth::token::builder::Fact;
+//! use biscuit_auth::builder::Fact;
 //!
 //! let f: Fact = "test(#data)".try_into().expect("parse error");
 //! ```
@@ -25,7 +25,7 @@ use nom::{
     error::{ErrorKind, FromExternalError, ParseError},
     multi::{many0, separated_list0, separated_list1},
     sequence::{delimited, pair, preceded, terminated, tuple},
-    IResult, Offset,
+    Finish, IResult, Offset,
 };
 use std::{
     collections::BTreeSet,
@@ -52,7 +52,7 @@ pub fn fact_inner(i: &str) -> IResult<&str, builder::Fact, Error> {
     let (i, fact_name) = name(i)?;
 
     let (i, _) = space0(i)?;
-    let (i, ids) = delimited(
+    let (i, terms) = delimited(
         char('('),
         cut(separated_list1(
             preceded(space0, char(',')),
@@ -61,13 +61,7 @@ pub fn fact_inner(i: &str) -> IResult<&str, builder::Fact, Error> {
         preceded(space0, char(')')),
     )(i)?;
 
-    Ok((
-        i,
-        builder::Fact(builder::Predicate {
-            name: fact_name.to_string(),
-            ids,
-        }),
-    ))
+    Ok((i, builder::Fact::new(fact_name.to_string(), terms)))
 }
 
 /// parse a Datalog check
@@ -76,11 +70,12 @@ pub fn check(i: &str) -> IResult<&str, builder::Check, Error> {
 
     let (i, _) = error(
         preceded(space0, eof),
-        |input| match input.chars().next() {
+        |input| {
+            match input.chars().next() {
             Some(')') => "unexpected parens".to_string(),
-            //Some('$') => "variables are not allowed in facts".to_string(),
             _ => format!("expected either the next term after ',' or the next check variant after 'or', but got '{}'",
                      input)
+        }
         },
         " ,\n",
     )(i)?;
@@ -99,15 +94,16 @@ fn check_inner(i: &str) -> IResult<&str, builder::Check, Error> {
 
 /// parse an allow or deny rule
 pub fn policy(i: &str) -> IResult<&str, builder::Policy, Error> {
-    let(i, policy) = policy_inner(i)?;
+    let (i, policy) = policy_inner(i)?;
 
     let (i, _) = error(
         preceded(space0, eof),
-        |input| match input.chars().next() {
+        |input| {
+            match input.chars().next() {
             Some(')') => "unexpected parens".to_string(),
-            //Some('$') => "variables are not allowed in facts".to_string(),
             _ => format!("expected either the next term after ',' or the next policy variant after 'or', but got '{}'",
                      input)
+        }
         },
         " ,\n",
     )(i)?;
@@ -161,10 +157,10 @@ pub fn check_body(i: &str) -> IResult<&str, Vec<builder::Rule>, Error> {
     let queries = queries
         .drain(..)
         .map(|rule_body| {
-            builder::Rule(
+            builder::Rule::new(
                 builder::Predicate {
                     name: "query".to_string(),
-                    ids: Vec::new(),
+                    terms: Vec::new(),
                 },
                 rule_body.0,
                 rule_body.1,
@@ -182,9 +178,10 @@ pub fn rule(i: &str) -> IResult<&str, builder::Rule, Error> {
         preceded(space0, eof),
         |input| match input.chars().next() {
             Some(')') => "unexpected parens".to_string(),
-            //Some('$') => "variables are not allowed in facts".to_string(),
-            _ => format!("expected the next term or expression after ',', but got '{}'",
-                     input)
+            _ => format!(
+                "expected the next term or expression after ',', but got '{}'",
+                input
+            ),
         },
         " ,\n",
     )(i)?;
@@ -198,16 +195,16 @@ pub fn rule_inner(i: &str) -> IResult<&str, builder::Rule, Error> {
 
     let (i, _) = tag("<-")(i)?;
 
-    let (i, (predicates, expressions)) = cut(rule_body)(i)?;
+    let (i, (body, expressions)) = cut(rule_body)(i)?;
 
-    let rule = builder::Rule(head, predicates, expressions);
+    let rule = builder::Rule::new(head, body, expressions);
 
     if let Err(message) = rule.validate_variables() {
-        return Err( nom::Err::Error(Error {
-                input: head_input,
-                code: ErrorKind::Satisfy,
-                message: Some(message),
-            }))
+        return Err(nom::Err::Failure(Error {
+            input: head_input,
+            code: ErrorKind::Satisfy,
+            message: Some(message),
+        }));
     }
 
     Ok((i, rule))
@@ -217,9 +214,7 @@ impl TryFrom<&str> for builder::Fact {
     type Error = error::Token;
 
     fn try_from(value: &str) -> Result<Self, Self::Error> {
-        fact(value)
-            .map(|(_, o)| o)
-            .map_err(|_| error::Token::ParseError)
+        Ok(fact(value).finish().map(|(_, o)| o)?)
     }
 }
 
@@ -227,12 +222,7 @@ impl TryFrom<&str> for builder::Rule {
     type Error = error::Token;
 
     fn try_from(value: &str) -> Result<Self, Self::Error> {
-        rule(value)
-            .map(|(_, o)| o)
-            .map_err(|e| {
-                println!("rule parsing error: {:?}", e);
-                error::Token::ParseError
-            })
+        Ok(rule(value).finish().map(|(_, o)| o)?)
     }
 }
 
@@ -240,9 +230,7 @@ impl FromStr for builder::Fact {
     type Err = error::Token;
 
     fn from_str(s: &str) -> Result<Self, Self::Err> {
-        fact(s)
-            .map(|(_, o)| o)
-            .map_err(|_| error::Token::ParseError)
+        Ok(fact(s).finish().map(|(_, o)| o)?)
     }
 }
 
@@ -250,9 +238,7 @@ impl FromStr for builder::Rule {
     type Err = error::Token;
 
     fn from_str(s: &str) -> Result<Self, Self::Err> {
-        rule(s)
-            .map(|(_, o)| o)
-            .map_err(|_| error::Token::ParseError)
+        Ok(rule(s).finish().map(|(_, o)| o)?)
     }
 }
 
@@ -260,9 +246,7 @@ impl TryFrom<&str> for builder::Check {
     type Error = error::Token;
 
     fn try_from(value: &str) -> Result<Self, Self::Error> {
-        check(value)
-            .map(|(_, o)| o)
-            .map_err(|_| error::Token::ParseError)
+        Ok(check(value).finish().map(|(_, o)| o)?)
     }
 }
 
@@ -270,9 +254,7 @@ impl FromStr for builder::Check {
     type Err = error::Token;
 
     fn from_str(s: &str) -> Result<Self, Self::Err> {
-        check(s)
-            .map(|(_, o)| o)
-            .map_err(|_| error::Token::ParseError)
+        Ok(check(s).finish().map(|(_, o)| o)?)
     }
 }
 
@@ -280,9 +262,7 @@ impl TryFrom<&str> for builder::Policy {
     type Error = error::Token;
 
     fn try_from(value: &str) -> Result<Self, Self::Error> {
-        policy(value)
-            .map(|(_, o)| o)
-            .map_err(|_| error::Token::ParseError)
+        Ok(policy(value).finish().map(|(_, o)| o)?)
     }
 }
 
@@ -290,9 +270,7 @@ impl FromStr for builder::Policy {
     type Err = error::Token;
 
     fn from_str(s: &str) -> Result<Self, Self::Err> {
-        policy(s)
-            .map(|(_, o)| o)
-            .map_err(|_| error::Token::ParseError)
+        Ok(policy(s).finish().map(|(_, o)| o)?)
     }
 }
 
@@ -300,9 +278,23 @@ impl FromStr for builder::Predicate {
     type Err = error::Token;
 
     fn from_str(s: &str) -> Result<Self, Self::Err> {
-        predicate(s)
-            .map(|(_, o)| o)
-            .map_err(|_| error::Token::ParseError)
+        Ok(predicate(s).finish().map(|(_, o)| o)?)
+    }
+}
+
+impl TryFrom<&str> for builder::BlockBuilder {
+    type Error = error::Token;
+
+    fn try_from(value: &str) -> Result<Self, Self::Error> {
+        match parse_block_source(value) {
+            Ok(mut result) => Ok(builder::BlockBuilder {
+                facts: result.facts.drain(..).map(|(_, fact)| fact).collect(),
+                rules: result.rules.drain(..).map(|(_, rule)| rule).collect(),
+                checks: result.checks.drain(..).map(|(_, check)| check).collect(),
+                context: None,
+            }),
+            Err(e) => Err(e.into()),
+        }
     }
 }
 
@@ -311,7 +303,7 @@ fn predicate(i: &str) -> IResult<&str, builder::Predicate, Error> {
     let (i, fact_name) = name(i)?;
 
     let (i, _) = space0(i)?;
-    let (i, ids) = delimited(
+    let (i, terms) = delimited(
         char('('),
         cut(separated_list1(preceded(space0, char(',')), cut(term))),
         preceded(space0, char(')')),
@@ -321,7 +313,7 @@ fn predicate(i: &str) -> IResult<&str, builder::Predicate, Error> {
         i,
         builder::Predicate {
             name: fact_name.to_string(),
-            ids,
+            terms,
         },
     ))
 }
@@ -331,7 +323,7 @@ fn rule_head(i: &str) -> IResult<&str, builder::Predicate, Error> {
     let (i, fact_name) = name(i)?;
 
     let (i, _) = space0(i)?;
-    let (i, ids) = delimited(
+    let (i, terms) = delimited(
         char('('),
         cut(separated_list0(preceded(space0, char(',')), cut(term))),
         preceded(space0, char(')')),
@@ -341,7 +333,7 @@ fn rule_head(i: &str) -> IResult<&str, builder::Predicate, Error> {
         i,
         builder::Predicate {
             name: fact_name.to_string(),
-            ids,
+            terms,
         },
     ))
 }
@@ -490,6 +482,8 @@ fn binary_op_4(i: &str) -> IResult<&str, builder::Binary, Error> {
         value(Binary::Prefix, tag("starts_with")),
         value(Binary::Suffix, tag("ends_with")),
         value(Binary::Regex, tag("matches")),
+        value(Binary::Intersection, tag("intersection")),
+        value(Binary::Union, tag("union")),
     ))(i)
 }
 
@@ -558,7 +552,7 @@ fn expr4(i: &str) -> IResult<&str, Expr, Error> {
 }
 
 fn name(i: &str) -> IResult<&str, &str, Error> {
-    let is_name_char = |c: char| is_alphanumeric(c as u8) || c == '_';
+    let is_name_char = |c: char| is_alphanumeric(c as u8) || c == '_' || c == ':';
 
     reduce(take_while1(is_name_char), " ,:(\n;")(i)
 }
@@ -607,7 +601,7 @@ fn parse_date(i: &str) -> IResult<&str, u64, Error> {
     map_res(
         map_res(
             take_while1(|c: char| c != ',' && c != ' ' && c != ')' && c != ']' && c != ';'),
-            |s| chrono::DateTime::parse_from_rfc3339(s),
+            chrono::DateTime::parse_from_rfc3339,
         ),
         |t| t.timestamp().try_into(),
     )(i)
@@ -659,7 +653,6 @@ fn set(i: &str) -> IResult<&str, builder::Term, Error> {
     let mut kind: Option<u8> = None;
     for term in list.drain(..) {
         let index = match term {
-            builder::Term::Symbol(_) => 0,
             builder::Term::Variable(_) => panic!("variables are not permitted in sets"),
             builder::Term::Integer(_) => 2,
             builder::Term::Str(_) => 3,
@@ -696,10 +689,9 @@ fn term_in_fact(i: &str) -> IResult<&str, builder::Term, Error> {
     preceded(
         space0,
         error(
-            alt((symbol, string, date, integer, bytes, boolean, set)),
+            alt((symbol, string, date, variable, integer, bytes, boolean, set)),
             |input| match input.chars().next() {
                 None | Some(',') | Some(')') => "missing term".to_string(),
-                Some('$') => "variables are not allowed in facts".to_string(),
                 _ => "expected a valid term".to_string(),
             },
             " ,)\n;",
@@ -761,14 +753,14 @@ pub fn sep(i: &str) -> IResult<&str, &str, Error> {
     alt((tag(";"), eof))(i)
 }
 
-pub fn parse_source(mut i: &str) -> Result<(&str, SourceResult), Vec<Error>> {
+pub fn parse_source(mut i: &str) -> Result<SourceResult, Vec<Error>> {
     let mut result = SourceResult::default();
     let mut errors = Vec::new();
 
     loop {
         if i.is_empty() {
             if errors.is_empty() {
-                return Ok((i, result));
+                return Ok(result);
             } else {
                 return Err(errors);
             }
@@ -838,14 +830,14 @@ pub fn parse_source(mut i: &str) -> Result<(&str, SourceResult), Vec<Error>> {
     }
 }
 
-pub fn parse_block_source(mut i: &str) -> Result<(&str, SourceResult), Vec<Error>> {
+pub fn parse_block_source(mut i: &str) -> Result<SourceResult, Vec<Error>> {
     let mut result = SourceResult::default();
     let mut errors = Vec::new();
 
     loop {
         if i.is_empty() {
             if errors.is_empty() {
-                return Ok((i, result));
+                return Ok(result);
             } else {
                 return Err(errors);
             }
@@ -873,7 +865,7 @@ pub fn parse_block_source(mut i: &str) -> Result<(&str, SourceResult), Vec<Error
                     SourceElement::Fact(i, f) => result.facts.push((i, f)),
                     SourceElement::Rule(i, r) => result.rules.push((i, r)),
                     SourceElement::Check(i, c) => result.checks.push((i, c)),
-                    SourceElement::Policy(_, _) => {},
+                    SourceElement::Policy(_, _) => {}
                     SourceElement::Comment => {}
                 }
 
@@ -1011,9 +1003,9 @@ where
 
 #[cfg(test)]
 mod tests {
+    use super::Error;
     use crate::{datalog, token::builder};
     use nom::error::ErrorKind;
-    use super::Error;
 
     #[test]
     fn name() {
@@ -1057,7 +1049,7 @@ mod tests {
 
     #[test]
     fn constraint() {
-        use builder::{date, int, set, string, symbol, var, Binary, Op, Unary};
+        use builder::{date, int, set, string, var, Binary, Op, Unary};
         use std::collections::BTreeSet;
         use std::time::{Duration, SystemTime};
 
@@ -1252,12 +1244,12 @@ mod tests {
             ))
         );
 
-        let h = [symbol("abc"), symbol("def")]
+        let h = [string("abc"), string("def")]
             .iter()
             .cloned()
             .collect::<BTreeSet<_>>();
         assert_eq!(
-            super::expr("[#abc, #def].contains($0)").map(|(i, o)| (i, o.opcodes())),
+            super::expr("[\"abc\", \"def\"].contains($0)").map(|(i, o)| (i, o.opcodes())),
             Ok((
                 "",
                 vec![
@@ -1269,7 +1261,7 @@ mod tests {
         );
 
         assert_eq!(
-            super::expr("![#abc, #def].contains($0)").map(|(i, o)| (i, o.opcodes())),
+            super::expr("![\"abc\", \"def\"].contains($0)").map(|(i, o)| (i, o.opcodes())),
             Ok((
                 "",
                 vec![
@@ -1298,17 +1290,6 @@ mod tests {
                 )
             ))
         );
-
-        use nom::error::ErrorKind;
-        // facts should not contain variables
-        assert_eq!(
-            super::fact("right( #authority, $var, #read )"),
-            Err(nom::Err::Failure(super::Error {
-                code: ErrorKind::Char,
-                input: "$var",
-                message: Some("variables are not allowed in facts".to_string()),
-            }))
-        );
     }
 
     #[test]
@@ -1411,7 +1392,7 @@ mod tests {
     fn rule_with_unused_head_variables() {
         assert_eq!(
             super::rule("right(#authority, $0, $test) <- resource( #ambienu, $0), operation(#ambient, #read)"),
-            Err( nom::Err::Error(Error {
+            Err( nom::Err::Failure(Error {
                 input: "right(#authority, $0, $test)",
                 code: ErrorKind::Satisfy,
                 message: Some("rule head contains variables that are not used in predicates of the rule's body: $test".to_string()),
@@ -1610,7 +1591,7 @@ mod tests {
         let printed = e.print(&syms).unwrap();
         println!("print: {}", e.print(&syms).unwrap());
         let h = HashMap::new();
-        let result = e.evaluate(&h).unwrap();
+        let result = e.evaluate(&h, &syms).unwrap();
         println!("evaluates to: {:?}", result);
 
         assert_eq!(
@@ -1624,7 +1605,7 @@ mod tests {
             ]
         );
         assert_eq!(&printed, "1 + 2 * 3");
-        assert_eq!(result, datalog::ID::Integer(7));
+        assert_eq!(result, datalog::Term::Integer(7));
 
         let input = " (1 + 2) * 3 ";
         println!("parsing: {}", input);
@@ -1637,7 +1618,7 @@ mod tests {
         let printed = e.print(&syms).unwrap();
         println!("print: {}", e.print(&syms).unwrap());
         let h = HashMap::new();
-        let result = e.evaluate(&h).unwrap();
+        let result = e.evaluate(&h, &syms).unwrap();
         println!("evaluates to: {:?}", result);
 
         assert_eq!(
@@ -1652,7 +1633,7 @@ mod tests {
             ]
         );
         assert_eq!(&printed, "(1 + 2) * 3");
-        assert_eq!(result, datalog::ID::Integer(9));
+        assert_eq!(result, datalog::Term::Integer(9));
     }
 
     #[test]
@@ -1785,8 +1766,7 @@ mod tests {
             },
         ];
 
-        let (_remaining, mut result) = res.unwrap();
-        //assert_eq!(remaining, "\n");
+        let mut result = res.unwrap();
         assert_eq!(
             result.facts.drain(..).map(|(_, r)| r).collect::<Vec<_>>(),
             expected_facts
@@ -1912,8 +1892,7 @@ mod tests {
             },
         ];
 
-        let (_remaining, mut result) = res.unwrap();
-        //assert_eq!(remaining, "\n");
+        let mut result = res.unwrap();
         assert_eq!(
             result.facts.drain(..).map(|(_, r)| r).collect::<Vec<_>>(),
             expected_facts
